-----------------------------------------------------------------------
--
--  File:        console_io.adb
--  Description: Standard console I/O package for Windows
--  Rev:         $Revision$
--  Date:        $Date$
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 2003
--  Billie Holidaystraat 28
--  2334 LK Leiden
--  The Netherlands
--  tel int +33 (0)71 531 4365
--
--  This is free software; you can redistribute it and/or modify it
--  under terms of the GNU General Public License as published by the
--  Free Software Foundation; either version 2, or (at your option) any
--  later version. This software is distributed in the hope that it will
--  be useful, but WITHOUT ANY WARRANTY; without even the implied
--  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
--  the GNU General Public License for more details. You should have
--  received a copy of the GNU General Public License distributed with
--  this software; see the file COPYING. If not, write to the Free
--  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
--  02111-1307, USA.
--
--  As a special exception, if other files instantiate generics from
--  this software, or you link this software with other files to
--  produce an executable, this software does not by itself cause the
--  resulting executable to be covered by the GNU General Public
--  License. This exception does not however invalidate any other
--  reasons why the executable file might be covered by the GNU Public
--  License.
--
--  Changelog:
--
--  $Log$
--
-----------------------------------------------------------------------

pragma C_Pass_By_Copy (128);

with System;

package body Console_IO is

   ----------------[ Win32 Binding ]----------------

   --  base types

   type NIBBLE is mod 2 ** 4;
   for NIBBLE'Size use 4;

   type BYTE is mod 2 ** 8;
   for BYTE'Size use 8;

   type WORD is mod 2 ** 16;
   for WORD'Size use 16;

   type DWORD is mod 2 ** 32;
   for DWORD'Size use 32;

   type HANDLE is new DWORD;

   --  structured types

   type Attribute is
      record
         Foreground : NIBBLE;
         Background : NIBBLE;
         Reserved   : BYTE := 0;
      end record;

   for Attribute use
      record
         Foreground at 0 range 0 .. 3;
         Background at 0 range 4 .. 7;
         Reserved   at 1 range 0 .. 7;
      end record;

   for Attribute'Size use 16;
   pragma Convention (C, Attribute);

   type COORD is
      record
         X : WORD;
         Y : WORD;
      end record;
   pragma Convention (C, COORD);

   type SMALL_RECT is
      record
         Left   : WORD;
         Top    : WORD;
         Right  : WORD;
         Bottom : WORD;
      end record;
   pragma Convention (C, SMALL_RECT);

   type CONSOLE_SCREEN_BUFFER_INFO is
      record
         Size       : COORD;
         Cursor_Pos : COORD;
         Attrib     : Attribute;
         Window     : SMALL_RECT;
         Max_Size   : COORD;
      end record;
   pragma Convention (C, CONSOLE_SCREEN_BUFFER_INFO);

   type CONSOLE_CURSOR_INFO is
      record
         Size    : DWORD;
         Visible : DWORD;
      end record;
   pragma Convention (C, CONSOLE_CURSOR_INFO);

   --  constants

   ENABLE_LINE_INPUT    : constant DWORD  := 16#00000002#;
   ENABLE_ECHO_INPUT    : constant DWORD  := 16#00000004#;
   INVALID_HANDLE_VALUE : constant HANDLE := 16#FFFFFFFF#;
   STD_INPUT_HANDLE     : constant DWORD  := 16#FFFFFFF6#;
   STD_OUTPUT_HANDLE    : constant DWORD  := 16#FFFFFFF5#;

   Color_Value : constant array (Color_Type) of NIBBLE :=
     (0, 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15);

   Color_Type_Value : constant array (NIBBLE) of Color_Type :=
     (Black, Blue, Green, Cyan, Red, Magenta, Brown, Gray,
      Black, Light_Blue, Light_Green, Light_Cyan, Light_Red,
      Light_Magenta, Yellow, White);

   --  API imports from msvcrt.dll (or crtdll.dll)

   function Get_Ch return Integer;
   pragma Import (C, Get_Ch, "_getch");

   function Kb_Hit return Integer;
   pragma Import (C, Kb_Hit, "_kbhit");

   --  API imports from kernel32.dll

   function Beep (Freqency, Duration : DWORD) return DWORD;
   pragma Import (StdCall, Beep, "Beep");

   function FillConsoleOutputAttribute
     (hConsoleOutput : HANDLE;
      Attrib         : Attribute;
      Length         : DWORD;
      Start          : COORD;
      Written        : System.Address) return DWORD;
   pragma Import (StdCall,
                  FillConsoleOutputAttribute,
                  "FillConsoleOutputAttribute");

   function FillConsoleOutputCharacter
     (hConsoleOutput : HANDLE;
      Char           : Character;
      Length         : DWORD;
      Start          : COORD;
      Written        : System.Address) return DWORD;
   pragma Import (StdCall,
                  FillConsoleOutputCharacter,
                  "FillConsoleOutputCharacterA");

   function GetConsoleCursorInfo
     (hConsoleOutput : HANDLE;
      Cursor         : System.Address) return DWORD;
   pragma Import (StdCall, GetConsoleCursorInfo, "GetConsoleCursorInfo");

   function GetConsoleMode
     (hConsoleOutput : HANDLE;
      Mode           : System.Address) return DWORD;
   pragma Import (StdCall, GetConsoleMode, "GetConsoleMode");

   function GetConsoleScreenBufferInfo
     (hConsoleOutput : HANDLE;
      ScreenInfo     : System.Address) return DWORD;
   pragma Import (StdCall,
                  GetConsoleScreenBufferInfo,
                  "GetConsoleScreenBufferInfo");

   function GetConsoleTitle
     (lpConsoleTitle : System.Address;
      nSize          : DWORD) return DWORD;
   pragma Import (StdCall, GetConsoleTitle, "GetConsoleTitleA");

   function GetStdHandle (Kind : DWORD) return HANDLE;
   pragma Import (StdCall, GetStdHandle, "GetStdHandle");

   function ReadFile
     (hConsoleInput        : HANDLE;
      lpBuffer             : System.Address;
      nNumberOfCharsToRead : DWORD;
      lpNumberOfCharsRead  : System.Address;
      lpReserved           : DWORD) return DWORD;
   pragma Import (StdCall, ReadFile, "ReadFile");

   function SetConsoleCursorInfo
     (hConsoleOutput : HANDLE;
      Cursor         : System.Address) return DWORD;
   pragma Import (StdCall, SetConsoleCursorInfo, "SetConsoleCursorInfo");

   function SetConsoleCursorPosition
     (hConsoleOutput : HANDLE;
      Position       : COORD) return DWORD;
   pragma Import (StdCall,
                  SetConsoleCursorPosition,
                  "SetConsoleCursorPosition");

   function SetConsoleMode
     (hConsoleOutput : HANDLE;
      Mode           : DWORD) return DWORD;
   pragma Import (StdCall, SetConsoleMode, "SetConsoleMode");

   function SetConsoleTextAttribute
     (hConsoleOutput : HANDLE;
      Attr           : Attribute) return DWORD;
   pragma Import (StdCall, SetConsoleTextAttribute, "SetConsoleTextAttribute");

   function SetConsoleTitle (lpConsoleTitle : System.Address) return DWORD;
   pragma Import (StdCall, SetConsoleTitle, "SetConsoleTitleA");

   function WriteFile
     (hConsoleinput          : HANDLE;
      lpBuffer               : System.Address;
      nNumberOfCharsToWrite  : DWORD;
      lpNumberOfCharsWritten : System.Address;
      lpReserved             : DWORD) return DWORD;
   pragma Import (StdCall, WriteFile, "WriteFile");

   --  exceptions

   Console_IO_Error : exception;
   --  bail out exception (should never occur)

   -----------------------------
   -- Package state variables --
   -----------------------------

   --  these variables are initialized at package elaboration

   hStdin  : HANDLE;                    --  Console input handle
   hStdout : HANDLE;                    --  Console output handle

   ----------------[ Local Declarations ]----------------

   --  constants

   Max_Buf_Len     : constant Integer   := 256; -- incl. CR/LF !
   Space_Character : constant Character := Character'Val (32);

   --  declarations

   function Unsigned_String (Item : String) return String;
   pragma Inline (Unsigned_String);

   --  implementation

   ---------------------
   -- Unsigned_String --
   ---------------------

   function Unsigned_String (Item : String) return String is
   begin
      if Item (Item'First) = Space_Character then
         return Item (Item'First + 1 .. Item'Last);
      else
         return Item;
      end if;
   end Unsigned_String;

   ----------------[ Implementation ]----------------
   package body Catenators is
   -------
   -- & --
   -------

   function "&" (L : Boolean; R : String) return String is
   begin
      return Boolean'Image (L) & R;
   end "&";

   function "&" (L : Duration; R : String) return String is
   begin
      return Unsigned_String (Duration'Image (L)) & R;
   end "&";

   function "&" (L : Float; R : String) return String is
   begin
      return Unsigned_String (Float'Image (L)) & R;
   end "&";

   function "&" (L : Integer; R : String) return String is
   begin
      return Unsigned_String (Integer'Image (L)) & R;
   end "&";

   function "&" (L : Long_Float; R : String) return String is
   begin
      return Unsigned_String (Long_Float'Image (L)) & R;
   end "&";

   function "&" (L : Long_Integer; R : String) return String is
   begin
      return Unsigned_String (Long_Integer'Image (L)) & R;
   end "&";

   function "&" (L : String; R : Boolean) return String is
   begin
      return L & Boolean'Image (R);
   end "&";

   function "&" (L : String; R : Duration) return String is
   begin
      return L & Unsigned_String (Duration'Image (R));
   end "&";

   function "&" (L : String; R : Float) return String is
   begin
      return L & Unsigned_String (Float'Image (R));
   end "&";

   function "&" (L : String; R : Integer) return String is
   begin
      return L & Unsigned_String (Integer'Image (R));
   end "&";

   function "&" (L : String; R : Long_Float) return String is
   begin
      return L & Unsigned_String (Long_Float'Image (R));
   end "&";

   function "&" (L : String; R : Long_Integer) return String is
   begin
      return L & Unsigned_String (Long_Integer'Image (R));
   end "&";

   end Catenators;
   
   ----------------------
   -- Background_Color --
   ----------------------

   function Background_Color return Color_Type is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Color_Type_Value (Screen.Attrib.Background);
   end Background_Color;

   procedure Background_Color (Color : Color_Type) is
      Attrib : Attribute;
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      Attrib.Background := Color_Value (Color);
      Attrib.Foreground := Screen.Attrib.Foreground;
      if SetConsoleTextAttribute (hStdout, Attrib) = 0 then
         raise Console_IO_Error;
      end if;
   end Background_Color;

   -----------
   -- Bleep --
   -----------

   procedure Bleep is
   begin
      if Beep (1000, 300) = 0 then
         raise Console_IO_Error;
      end if;
   end Bleep;

   ---------------
   -- Clear_Eol --
   ---------------

   procedure Clear_EOL is
      Written : DWORD;
      Length  : DWORD;
      Screen  : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      Length := DWORD (Screen.Size.X - Screen.Cursor_Pos.X);
      if FillConsoleOutputAttribute
        (hStdout,
         Screen.Attrib,
         Length,
         Screen.Cursor_Pos,
         Written'Address) = 0 then
         raise Console_IO_Error;
      end if;
      if FillConsoleOutputCharacter
        (hStdout,
         Space_Character,
         Length,
         Screen.Cursor_Pos,
         Written'Address) = 0 then
         raise Console_IO_Error;
      end if;
   end Clear_EOL;

   ------------------
   -- Clear_Screen --
   ------------------

   procedure Clear_Screen (Color : Color_Type := Black) is
      Bytes  : DWORD;
      Length : DWORD;
      Attrib : Attribute;
      Home   : COORD := (0, 0);
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      Length := DWORD (Screen.Size.X) * DWORD (Screen.Size.Y);
      Attrib.Background := Color_Value (Color);
      Attrib.Foreground := Screen.Attrib.Foreground;
      if SetConsoleTextAttribute (hStdout, Attrib) = 0 then
         raise Console_IO_Error;
      end if;
      if FillConsoleOutputAttribute
        (hStdout, Attrib, Length, Home, Bytes'Address) = 0 then
         raise Console_IO_Error;
      end if;
      if FillConsoleOutputCharacter
        (hStdout, Space_Character, Length, Home, Bytes'Address) = 0 then
         raise Console_IO_Error;
      end if;
      if SetConsoleCursorPosition (hStdout, Home) = 0 then
         raise Console_IO_Error;
      end if;
   end Clear_Screen;

   --------------------
   -- Cursor_Visible --
   --------------------

   function Cursor_Visible return Boolean is
      Cursor : CONSOLE_CURSOR_INFO;
   begin
      if GetConsoleCursorInfo (hStdout, Cursor'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Cursor.Visible = 1;
   end Cursor_Visible;

   procedure Cursor_Visible (Visible : Boolean) is
      Cursor : CONSOLE_CURSOR_INFO;
   begin
      if GetConsoleCursorInfo (hStdout, Cursor'Address) = 0 then
         raise Console_IO_Error;
      end if;
      if Visible then
         Cursor.Visible := 1;
      else
         Cursor.Visible := 0;
      end if;
      if SetConsoleCursorInfo (hStdout, Cursor'Address) = 0 then
         raise Console_IO_Error;
      end if;
   end Cursor_Visible;

   ----------------------
   -- Foreground_Color --
   ----------------------

   function Foreground_Color return Color_Type is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Color_Type_Value (Screen.Attrib.Foreground);
   end Foreground_Color;

   procedure Foreground_Color (Color : Color_Type) is
      Attrib : Attribute;
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      Attrib.Foreground := Color_Value (Color);
      Attrib.Background := Screen.Attrib.Background;
      if SetConsoleTextAttribute (hStdout, Attrib) = 0 then
         raise Console_IO_Error;
      end if;
   end Foreground_Color;

   -------------------
   -- Get_Character --
   -------------------

   function Get_Character return Character is
      Ch     : Character;
      Length : DWORD;
      Mode   : DWORD;
   begin
      if GetConsoleMode (hStdin, Mode'Address) = 0 then
         raise Console_IO_Error;
      end if;
      if SetConsoleMode
        (hStdin, Mode xor (ENABLE_LINE_INPUT or ENABLE_ECHO_INPUT)) = 0 then
         raise Console_IO_Error;
      end if;
      if ReadFile (hStdin, Ch'Address, 1, Length'Address, 0) = 0 then
         raise Console_IO_Error;
      end if;
      if SetConsoleMode (hStdin, Mode) = 0 then
         raise Console_IO_Error;
      end if;
      return Ch;
   end Get_Character;

   -------------
   -- Get_Key --
   -------------

   function Get_Key return Character is
      Ch : Integer := Get_Ch;
   begin
      if Ch = 16#00E0# then
         Ch := 0;
      end if;
      return Character'Val (Ch);
   end Get_Key;

   --------------
   -- Get_Line --
   --------------

   --  Boolean

   function Get_Line return Boolean is
      Input : constant String := Get_Line;
   begin
      return Boolean'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  Character

   function Get_Line return Character is
      Input : constant String := Get_Line;
   begin
      if Input'Length /= 1 then
         raise Input_Error;
      end if;
      return Input (Input'First);
   end Get_Line;

   --  Duration

   function Get_Line return Duration is
      Input : constant String := Get_Line;
   begin
      return Duration'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  Float

   function Get_Line return Float is
      Input : constant String := Get_Line;
   begin
      return Float'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  Integer

   function Get_Line return Integer is
      Input : constant String := Get_Line;
   begin
      return Integer'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  Long_Float

   function Get_Line return Long_Float is
      Input : constant String := Get_Line;
   begin
      return Long_Float'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  Long_Integer

   function Get_Line return Long_Integer is
      Input : constant String := Get_Line;
   begin
      return Long_Integer'Value (Input);
   exception
      when Constraint_Error => raise Input_Error;
   end Get_Line;

   --  String

   function Get_Line return String is
      Length : Integer;
      Ch     : Character;
      Buffer : String (1 .. Max_Buf_Len);
   begin

      --  read input
      if ReadFile
        (hStdin,
         Buffer'Address,
         DWORD (Max_Buf_Len),
         Length'Address, 0) = 0 then
         raise Console_IO_Error;
      end if;

      --  check for no input
      if Length = 2 and then Buffer (2) = ASCII.Lf then
         raise Input_Error;
      end if;

      --  check for too much input
      if Length = Max_Buf_Len and then Buffer (Max_Buf_Len) /= ASCII.Lf then
         --  skip input remaining in input buffer
         loop
            if ReadFile (hStdin, Ch'Address, 1, Length'Address, 0) = 0 then
               raise Console_IO_Error;
            end if;
            exit when Ch = ASCII.Lf;
         end loop;
         raise Input_Error;
      end if;

      --  return string
      return Buffer (1 .. Length - 2);

   end Get_Line;

   -------------
   -- Goto_XY --
   -------------

   procedure Goto_XY (X : Natural := 0; Y : Natural := 0) is
      Position : COORD := (WORD (X), WORD (Y));
   begin
      if (X >= Screen_Width) or (Y >= Screen_Height) then
         raise Cursor_Error;
      end if;
      if SetConsoleCursorPosition (hStdout, Position) = 0 then
         raise Console_IO_Error;
      end if;
   end Goto_XY;

   -------------------
   -- Key_Available --
   -------------------

   function Key_Available return Boolean is
   begin
      return Kb_Hit /= 0;
   end Key_Available;

   --------------
   -- New_Line --
   --------------

   procedure New_Line (Count : Positive := 1) is
   begin
      for I in 1 .. Count loop
         Put (ASCII.Lf);
      end loop;
   end New_Line;

   ------------
   -- Prompt --
   ------------

   --  Boolean

   function Prompt (Text : String) return Boolean is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Character

   function Prompt (Text : String) return Character is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Duration

   function Prompt (Text : String) return Duration is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Float

   function Prompt (Text : String) return Float is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Integer

   function Prompt (Text : String) return Integer is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Long_Float

   function Prompt (Text : String) return Long_Float is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  Long_Integer

   function Prompt (Text : String) return Long_Integer is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   --  String

   function Prompt (Text : String) return String is
   begin
      loop
         begin
            Put (Text & ": ");
            return Get_Line;
         exception
            when Input_Error => null;
         end;
      end loop;
   end Prompt;

   ---------
   -- Put --
   ---------

   --  Boolean

   procedure Put (Item : Boolean) is
   begin
      Put (Boolean'Image (Item));
   end Put;

   --  Character

   procedure Put (Item : Character) is
      Read : DWORD;
   begin
      if WriteFile (hStdout, Item'Address, 1, Read'Address, 0) = 0 then
         raise Console_IO_Error;
      end if;
   end Put;

   --  Duration

   procedure Put (Item : Duration; Signed : Boolean := False) is
      Value : constant String := Duration'Image (Item);
   begin
      if Signed then
         Put (Value);
      else
         Put (Unsigned_String (Value));
      end if;
   end Put;

   --  Float

   procedure Put (Item : Float; Signed : Boolean := False) is
      Value : constant String := Float'Image (Item);
   begin
      if Signed then
         Put (Value);
      else
         Put (Unsigned_String (Value));
      end if;
   end Put;

   --  Integer

   procedure Put (Item : Integer; Signed : Boolean := False) is
      Value : constant String := Integer'Image (Item);
   begin
      if Signed then
         Put (Value);
      else
         Put (Unsigned_String (Value));
      end if;
   end Put;

   --  Long_Float

   procedure Put (Item : Long_Float; Signed : Boolean := False) is
      Value : constant String := Long_Float'Image (Item);
   begin
      if Signed then
         Put (Value);
      else
         Put (Unsigned_String (Value));
      end if;
   end Put;

   --  Long_Integer

   procedure Put (Item : Long_Integer; Signed : Boolean := False) is
      Value : constant String := Long_Integer'Image (Item);
   begin
      if Signed then
         Put (Value);
      else
         Put (Unsigned_String (Value));
      end if;
   end Put;

   --  String

   procedure Put (Item : String) is
      Length : DWORD;
   begin
      if WriteFile
        (hStdout, Item'Address, Item'Length, Length'Address, 0) = 0 then
         raise Console_IO_Error;
      end if;
   end Put;

   --------------
   -- Put_Line --
   --------------

   --  Boolean

   procedure Put_Line (Item : Boolean) is
   begin
      Put_Line (Boolean'Image (Item));
   end Put_Line;

   --  Character

   procedure Put_Line (Item : Character) is
   begin
      Put (Item);
      New_Line;
   end Put_Line;

   --  Duration

   procedure Put_Line (Item : Duration; Signed : Boolean := False) is
   begin
      Put (Item, Signed);
      New_Line;
   end Put_Line;

   --  Float

   procedure Put_Line (Item : Float; Signed : Boolean := False) is
   begin
      Put (Item, Signed);
      New_Line;
   end Put_Line;

   --  Integer

   procedure Put_Line (Item : Integer; Signed : Boolean := False) is
   begin
      Put (Item, Signed);
      New_Line;
   end Put_Line;

   --  Long_Float

   procedure Put_Line (Item : Long_Float; Signed : Boolean := False) is
   begin
      Put (Item, Signed);
      New_Line;
   end Put_Line;

   --  Long_Integer

   procedure Put_Line (Item : Long_Integer; Signed : Boolean := False) is
   begin
      Put (Item, Signed);
      New_Line;
   end Put_Line;

   --  String

   procedure Put_Line (Item : String) is
   begin
      Put (Item);
      New_Line;
   end Put_Line;

   -------------------
   -- Screen_Height --
   -------------------

   function Screen_Height return Positive is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Positive (Screen.Window.Bottom - Screen.Window.Top + 1);
   end Screen_Height;

   ------------------
   -- Screen_Width --
   ------------------

   function Screen_Width return Positive is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Positive (Screen.Window.Right - Screen.Window.Left + 1);
   end Screen_Width;

   -----------
   -- Title --
   -----------

   function Title return String is
      Text   : String (1 .. Max_Buf_Len + 1);
      Result : DWORD;
      Length : constant DWORD := DWORD (Max_Buf_Len);
   begin
      Result := GetConsoleTitle (Text'Address, Length);
      if Result = 0 then
         raise Console_IO_Error;
      else
         return Text (1 .. Integer (Result));
      end if;
   end Title;

   procedure Title (Item : String) is
      Text : constant String := Item & Character'Val (0);
   begin
      if SetConsoleTitle (Text'Address) = 0 then
         raise Console_IO_Error;
      end if;
   end Title;

   -------------
   -- Where_X --
   -------------

   function Where_X return Natural is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Natural (Screen.Cursor_Pos.X);
   end Where_X;

   -------------
   -- Where_Y --
   -------------

   function Where_Y return Natural is
      Screen : CONSOLE_SCREEN_BUFFER_INFO;
   begin
      if GetConsoleScreenBufferInfo (hStdout, Screen'Address) = 0 then
         raise Console_IO_Error;
      end if;
      return Natural (Screen.Cursor_Pos.Y);
   end Where_Y;

begin

   --  get input handle
   hStdin := GetStdHandle (STD_INPUT_HANDLE);
   if hStdin = INVALID_HANDLE_VALUE then
      raise Console_IO_Error;
   end if;

   --  get output handle
   hStdout := GetStdHandle (STD_OUTPUT_HANDLE);
   if hStdout = INVALID_HANDLE_VALUE then
      raise Console_IO_Error;
   end if;

end Console_IO;
