with Ada.Finalization;	                                                                      --  1
use  Ada;			                                                                      --  2
generic				                                                                --  3
   type Element is tagged private;  -- A more robust design would define Element	          --  4
                                    --  as a derivation from Ada.Finalization.Controlled      --  5
   with function  Is_Valid(Data : Element) return Boolean;	                                  --  6
package Queue_Manager_1 is	                                                                --  7
   type List is limited private;                                                              --  8
   type List_Reference is access all List;                                                    --  9
   type List_Item is new Element with private;                                                -- 10
   type Item_Reference is access all List_Item'Class;                                         -- 11
   -- A classwide access type permitting a heterogenuous queue	                            -- 12
   procedure Clear (L : in out List);	                                                    -- 13
   procedure Insert_At_Head (L : in out List; I : in     List_Item'Class);                    -- 14		
   procedure Insert_At_Head (L : access List; I : access List_Item'Class);                    -- 15
   -- A more complete design would include added options for the Insert operation	          -- 16
   procedure Copy (Source : in List; Target : in out List);	                                  -- 17	
   function Remove_From_Tail (L : access List) return List_Item'Class;                        -- 18	
   -- A more complete design would include added options for the Remove operation	          -- 19
   function "=" (L, R : List) return Boolean;	                                              -- 20
   function  Node_Count (L : access List) return Natural;	                                  -- 21
   function  Is_Empty (L : access List) return Boolean;	                                  -- 22
   --	 ===============  Define the Active Iterator  =====================	                -- 23
                                                                                              -- 24
   type Iterator is private;                                                                  -- 25

   procedure Initialize_Iterator(This     : in out Iterator;	                            -- 26
                                 The_List : access List);	                                  -- 27
   function  Next(This : in Iterator) return Iterator; 	                                  -- 28
                                                                                              -- 29
   function  Get (This : in Iterator) return List_Item'Class;	                            -- 30
   function  Get (This : in Iterator) return Item_Reference;	                            -- 31
                                                                                              -- 32
   function  Is_Done(This : in Iterator) return Boolean;	                                  -- 33
                                                                                              -- 34
   Iterator_Error : exception;    	                                                          -- 35
private 				                                                                -- 36
   use Ada.Finalization;	                                                                -- 37
   type List_Node;                                                                            -- 38
   type Link is access all List_Node;                                                         -- 39
   type Iterator is new Link;                                                                 -- 40
   type List_Item is new Element with null record;                                            -- 41
   type List_Node is new Controlled with                                                      -- 42
      record                                                                                  -- 43                                                                          
         Data : Item_Reference;                                                               -- 44 
         Next : Link;                                                                         -- 45 
         Prev : Link;                                                                         -- 46 
      end record;                                                                             -- 47
   type List is new Limited_Controlled with                                                   -- 48
      record                                                                                  -- 49
         Count   : Natural := 0;                                                              -- 50 
         Head    : Link;                                                                      -- 51 
         Tail    : Link;                                                                      -- 52 
         Current : Link;                                                                      -- 53 
      end record;                                                                             -- 54
   procedure Finalize(One_Node : in out List_Node);	                                        -- 55
   procedure Finalize(The_List : in out List);	                                              -- 56
end Queue_Manager_1;	                                                                      -- 57